/* Loadmap.c - attempts to load a map file from duke nukem 3d and provide
some information about it

Now it loads from a file and displays it on screen.

Steve Bennett April-June 97 */

// $MEM (640K)

#include <dos.h>
#include <stdio.h>
#include <conio.h>
#include <io.h>
#include <stdlib.h>
#include <fcntl.h>
#include <alloc.h>
#include <graphics.h>
#include <assert.h>
#include <sys\stat.h>
#include <math.h>
#include <string.h>
#include <bios.h>

#define MAXSPRITELEVEL 5
#define MAXWALLLEVEL 2
#define SPRITEWIDTH 3
#define PI 3.141592653589

#define BLUEPAL 1
#define REDPAL 1
#define REDKEY 21
#define YELLOWKEY 23

#define LIGHTGREY LIGHTGRAY

#include "names.h"
#include "sbmouse.h"
#include "buildmap.h"
#include "nameh.h"

#pragma warn +wccc
#pragma warn -wsig
// *********************** STATS DECLARATION *********************** //
typedef struct {
	int numsectors, numwalls, numsprites;
	long lofloor, hifloor, loceil, hiceil;
	int brownwalls, twosided, maskedwalls, vidscreens;
	int blocksprites, flatsprites, floorsprites, vidcameras;
	int netgames, startpoints;
	int mirrors, tiltsectors;
	int fatsprite, thinsprite, tallsprite, shortsprite;
	int democams;
} statstype;

statstype Stats;

typedef enum { WALLS, SECTORS } Tmethod;

typedef struct { /* Various info on the 'screen' */
	long centrex, centrey;
	long scalex, scaley;
	int spritelevel;   /* What sort of sprites to display */
	int walllevel;   /* What sort of walls to display */
	int maxx, maxy;
	int Usinggraph;
	int clearmethod;  /* How to clear walls - individually or by screen */
	int spritewidth, spriteheight;
   int mapleft, maptop, mapbottom, mapright; /* Area of screen for map */
   int showspritedirs;
	int selectsprites;
	Tmethod drawmethod;
} Tscreen;

Tscreen Screen;

typedef struct {
	char solidwall, doublewall, mirror, oneway, maskable, blockingwall,
	enemy, item, rat, se, seenine, netgames, playerstart, wallflat, groundflat,
	blockingsprite, plainsprite,
	sector, teleporter, spritebox;
} Tcolors;

Tcolors Colors;

int Verbose = 0; // Prints out more debug info
int Dragmap = 0; // Map slides around when mouse is down
int mapfile;
int Cursprite = -1; // Which sprite the mouse is selecting

Telement * element;

int FileRead (int filehandle, void *buf, long numbytes) {
	static long Bytesread=0; // How many bytes read from file
	long bufread = 0;
	char *charbuf;  //char so we can access its size
	unsigned int i;
	long oread = Bytesread;

	charbuf = buf;
	for (i = 0; i <= numbytes / 32000l; i++) {
 		bufread = _read (filehandle, charbuf + i * 32000l,
			((numbytes - 32000l * i) > 32000) ? 32000 : numbytes % 32000l);
		Bytesread += (long)bufread;
	}
	if (Verbose)
		printf ("Tried to read %ld bytes, read %ld bytes in, total read is %ld\n", numbytes,
		Bytesread-oread, Bytesread);
	return bufread;
}

int FileWrite (int filehandle, void *buf, long numbytes) {
/* Writes a certain number of bytes from a place in memory to a file */
	long bufwrote;
	static long Byteswrote=0; // How many bytes read from file
	bufwrote = (unsigned)_write (filehandle, buf, numbytes);
	Byteswrote += (long)bufwrote;
	if (Verbose)
		printf ("Tried to write %d bytes, wrote %ld bytes out, total written is %ld\n",
			numbytes, bufwrote, Byteswrote);
	if (bufwrote < numbytes) {
		sound (100);
		delay (50);
		nosound();
	};
	return bufwrote;
}

/*char * Str (long n) {
char * S;
	S=malloc (log10(n));
	ltoa (n, S, 10);
	return S;
}*/

void far * FPAllocMem (long nobytes) {
unsigned seg;
	if (allocmem ((nobytes+15)/16, &seg) == -1)
		return MK_FP (seg, 0);
	else
		{ printf ("Uh-oh...not enough mem\n"); exit (1);
	}
	return 0;
}

int FPFreeMem (void far * p) {
unsigned seg;
	seg = FP_SEG (p);
	return freemem (seg);

}

/* called by the graphics kernel to allocate memory */
void far * far _graphgetmem(unsigned size)
{
		 if (Verbose) {
			 printf("_graphgetmem called to allocate %d bytes.\n", size);
			 printf("press any key:");
			 getch();
			 printf("\n");
		 }
		 /* allocate memory from far heap */
		 return FPAllocMem((long)size);
}

/* called by the graphics kernel to free memory */
void far _graphfreemem(void far *ptr, unsigned size)
{
		if (Verbose) {
			printf("_graphfreemem called to free %d bytes.\n", size);
			printf("press any key:");
			getch();
			printf("\n");
		}

		/* free ptr from far heap */
		FPFreeMem (ptr);
}



long ScreenXMapX (long screenx) {
/* Takes a screen position and calculates the map x from that */
	return (screenx - Screen.mapleft - (Screen.mapright-Screen.mapleft) / 2)
		* Screen.scalex + Screen.centrex;
}

long ScreenYMapY (long screeny) {
/* Takes a screen position and calculates the map y from that */
	return (screeny - Screen.maptop - (Screen.mapbottom - Screen.maptop) / 2)
		* Screen.scaley + Screen.centrey;
}

long MapXScreenX (long mapx) {
/* Takes a map location x and calculates the screen x from it */
	return ((mapx - Screen.centrex) / Screen.scalex) + (Screen.mapright-
	Screen.mapleft) / 2 + Screen.mapleft;
}


long MapYScreenY (long mapy) {
/* Takes a map location y and calculates the screen y from it */
	return ((mapy - Screen.centrey) / Screen.scaley) + (Screen.mapbottom-
	Screen.maptop) / 2 + Screen.maptop;
}

double Ang2Radians (short angle) {
/* Converts BUILD angles from 0-2047 to radians from 0-(2 * PI) */
	return (2048 - angle) / (double) 2048 * PI * 2;
}

char * GetSEName (int setype) {
/* Returns the name of type of Sector Effector */
	switch (setype) {
		case 0: return "Rotating Sector";
		case 1: return "Pivot point";
		case 2: return "Earthquake";
		case 3: return "Lights flicker/shot out";
		case 4: return "Lights flicker";
		case 5: return "SE type 5";
		case 6: return "Subway";
		case 7: return "Teleporter";
		case 8: return "Lights/door up";
		case 9: return "Lights/door down";
		case 10: return "Door auto-shut";
		case 11: return "Swing door";
		case 12: return "Light switch";
		case 13: return "Explosive";
		case 14: return "Subway carriage";
		case 15: return "Sliding door";
		case 16: return "Roate reactor";
		case 17: return "Elevator teleport";
		case 18: return "SE type 18";
		case 19: return "Ceiling Down/shot";
		case 20: return "Bridge";
		case 21: return "Drop floor";
		case 22: return "Prong door";
		case 23: return "SE type 23";
		case 24: return "Conveyor belt";
		case 25: return "Engine";
		case 26: return "SE type 26";
		case 27: return "Demo camera";
		case 28: return "SE type 28";
		case 29: return "Float (water)";
		case 30: return "2-way train";
		case 31: return "Floor rises";
		case 32: return "Ceiling falls";
		case 33: return "Jibs/Quake";
		case 34: return "SE type 34";
		case 35: return "SE type 35";
		case 36: return "Shrink ray";
	}
	return "[Unknown]";
}


long GetSpriteNum (long mapx, long mapy, Tmap *map) {
/* Looks through all the sprites and finds a sprite with the co-ords given */
long i;
int range=100;
int ret = -1;
	for (i=0; i <= map -> numsprites; i++) {
		if (abs (map->sprite [i].x - mapx) <= range &&
			 abs (map->sprite [i].y - mapy) <= range) {
				ret = i;
				i = map -> numsprites + 1;
			}
	}
	return ret;
}

/* Does memory allocation etc to have one more sprite around */
void InsertSprite (Tmap * map, Tsprite * sprite) {
	map -> sprite = realloc (map->sprite, sizeof (Tsprite)); //##
	memcpy (map->sprite + map->numsprites, sprite, sizeof (Tsprite));//!
	map->numsprites++;
}

/* Makes a sprite exist on the map, with a few fields set */
void CreateSprite (Tmap * map, int tilenum, unsigned lotag, unsigned hitag,
	long mapx, long mapy) {
Tsprite sprite;

	sprite.picnum = tilenum;
	sprite.lotag = lotag;
	sprite.hitag = hitag;
	sprite.x = mapx;
   sprite.y = mapy;

	InsertSprite (map, &sprite);
	memcpy (map->sprite + map->numsprites-1, &sprite, sizeof (Tsprite));
}

void HiLiteSprite (Tsprite *sprite) {
/* Makes a sprite on the map stand out by enclosing by a square */
long swid, shei;
int sx, sy;
	setlinestyle (0, 0, 3);

	swid = Screen.spritewidth * 1.5 * sprite->xrepeat/100;
	shei = Screen.spritewidth * 1.5 * sprite->yrepeat/100;
	sx = MapXScreenX (sprite->x);
	sy = MapYScreenY (sprite->y);
	setcolor (Colors.spritebox);
	rectangle (sx-swid, sy-shei, sx+swid, sy+shei);
}

void UnHiLiteSprite (Tsprite *sprite) {
/* Removes the red box around a selected sprite */
long swid, shei;
int sx, sy;
	setlinestyle (0, 0, 3);

	swid = Screen.spritewidth * 1.5 * sprite->xrepeat/100;
	shei = Screen.spritewidth * 1.5 * sprite->yrepeat/100;
	sx = MapXScreenX (sprite->x);
	sy = MapYScreenY (sprite->y);
	setcolor (0);
	rectangle (sx-swid, sy-shei, sx+swid, sy+shei);
}


void ShowPlayerStart (Tplayerstart *start, int clearing){
/* Displays the point on the map where player one begins his journey.
Represented by a purple asterisk */
long swid, shei;
int sx, sy, dx, dy;
double ang;
	swid = Screen.spritewidth * 1.5;
	shei = Screen.spriteheight * 1.5;
	sx = MapXScreenX (start->posx);
	sy = MapYScreenY (start->posy);
	setlinestyle (USERBIT_LINE, 0xfefe, 1);
	setcolor (5 * (1 - clearing));
	line (sx-swid, sy-shei, sx+swid, sy+shei); /* These 4 lines draw asterisk*/
	line (sx+swid, sy-shei, sx-swid, sy+shei);
	line (sx-swid, sy, sx+swid, sy);
	line (sx, sy-shei, sx, sy+shei);

	ang = Ang2Radians (start -> angle);

	dx = (int)(swid * cos (ang));
	dy = (int)(shei * sin (ang));

	setcolor (15);
	setlinestyle (0, 0, 0);
	line (sx, sy, sx+dx, sy-dy);

}


void ZoomOut (void) {
/* Changes scale factors, zooming out.  Still have to redraw screen */
	Screen.scalex = Screen.scalex * 6/5;
	Screen.scaley = Screen.scaley * 6/5;
	Screen.spritewidth = SPRITEWIDTH * 100 / Screen.scalex;
	Screen.spriteheight = SPRITEWIDTH * 100 / Screen.scaley;
}

void ZoomIn (void) {
/* Changes scale factors, zooming in.  Still have to redraw screen */
	if (Screen.scalex>=3) {  // Prevents over-zooming in which gives /0
		Screen.scalex = Screen.scalex * 5/6;
		Screen.scaley = Screen.scaley * 5/6;
		Screen.spritewidth = SPRITEWIDTH * 100 / Screen.scalex;
		Screen.spriteheight = SPRITEWIDTH * 100 / Screen.scaley;
	}
}

void ShowSectorTxt (Tsector *sector) {
	printf ("wallptr: %d ", sector->wallptr);
	printf ("wallnum: %d ", sector->wallnum);
	printf ("ceilingz: %ld ", sector->ceilingz);
	printf ("floorz: %ld ", sector->floorz);
	printf ("ceilingstat: %d ", sector->ceilingstat);
	printf ("floorstat: %d ", sector->floorstat);
	printf ("ceilingpicnum: %d ", sector->ceilingpicnum);
	printf ("ceilingheinum: %d ", sector->ceilingheinum);
	printf ("ceilingshade %d ", sector->ceilingshade);
	printf ("ceilingpal %d ", sector->ceilingpal);
	printf ("ceilingxpanning %d ", sector->ceilingxpanning);
	printf ("ceilingypanning %d ", sector->ceilingypanning);
	printf ("floorpicnum %d ", sector->floorpicnum);
	printf ("floorheinum %d ", sector->floorheinum);
	printf ("floorshade %d ", sector->floorshade);
	printf ("floorpal: %d floorxpanning: %d ", sector->floorxpanning, sector->floorypanning);
	printf ("visibility: %d ", sector->visibility);
	printf ("filler: % d ", sector->filler);
	printf ("lotag: %d hitag: %d ", sector->lotag, sector->hitag);
	printf ("extra: %d ", sector->extra);

}

void SectorStats (Tsector *sector) {

	if (abs (sector->floorz/100) < 5000 && abs (sector->ceilingz/100 < 5000) ){
		if (sector->  floorz > Stats.hifloor) Stats.hifloor = sector->floorz;
		if (sector->  floorz < Stats.lofloor) Stats.lofloor = sector->floorz;
		if (sector->ceilingz > Stats.hiceil ) Stats.hiceil  = sector->ceilingz;
		if (sector->ceilingz < Stats.loceil ) Stats.loceil  = sector->ceilingz;
		if (sector->ceilingstat.sloped || sector->floorstat.sloped)
			Stats.tiltsectors++;

	}
	else
	{
	 printf("** Out of range **");
	}
}

void ShowWallTxt (Twall *wall) {
	printf ("x,y :%ld,%ld ", wall->x, wall->y);
	printf ("point2 :%d nextwall: %d nextsector: %d ", wall->point2,
		wall->nextwall, wall->nextsector);
	printf ("cstat :%d ", wall->cstat);
	printf ("picnum :%d ", wall->picnum);
	printf ("overpicnum :%d ", wall->overpicnum);
	printf ("shade :%d ", wall->shade);
	printf ("pal :%d xrepeat %d yrepeat %d xpanning %d ypanning %d ",
		wall->pal, wall->xrepeat, wall->yrepeat, wall->xpanning, wall->ypanning);
	printf ("lotag: %d ", wall->lotag);
	printf ("hitag :%d ", wall->hitag);
	printf ("extra :%d ", wall->extra);
}

void ShowWall (Twall *wall, Twall *wall2, int clearing) {
long sx1, sy1, sx2, sy2;
int col, dashed = 0;

	/* work out screen locations from map pos's*/
	sx1 = MapXScreenX (wall -> x);
	sy1 = MapYScreenY (wall -> y);
	sx2 = MapXScreenX (wall2 -> x);
	sy2 = MapYScreenY (wall2 -> y);

	if (sx1 > Screen.mapleft && sx1 < Screen.mapright && sx2 > Screen.mapleft
		&& sx2 < Screen.mapright && sy1 > Screen.maptop && sy1 < Screen.mapbottom
		&& sy2 > Screen.maptop && sy2 < Screen.mapbottom) {

		col = 14;    // Defaults to yellow...if you see a yellow wall, mistake!

		if (wall->nextsector == -1) // Solid walls are always white
			col = Colors.solidwall;
		else if (wall->overpicnum==560) // Mirrors are always green
			col = Colors.mirror;
		else if (wall->cstat.masked) // General masked walls (glass etc) blue
			col = Colors.maskable;
		else if (wall->cstat.blocking) // Solid (blocking) walls bt purple
			col = Colors.blockingsprite;
		else
			col = Colors.doublewall;	 // Other two-sided walls red.

		if (clearing == 1) col = 0;
		if (wall->cstat.oneway) {
			dashed = 1;
			col = Colors.oneway;
			}
		if (clearing == 2)
			col = Colors.sector;

		setcolor (col);

		if (wall->cstat.hitscan)       // Hitscan (shootable) walls thick
			setlinestyle (dashed * DOTTED_LINE,  0, 3);
		else
			setlinestyle (dashed * DOTTED_LINE,  0, 1);

		line (sx1, sy1, sx2, sy2);
	}
}

void ShowSector (Tsector *sector, Twall *walls)
{
	int i;
	Twall *wp;
	wp = walls + sector->wallptr;
	for (i = 0; i < sector->wallnum; i++) {
		ShowWall (wp, walls+(wp -> point2), 2);
		wp++;
	}

};

void HideSector (Tsector *sector, Twall *walls)
{
	int i;
	Twall *wp;
	wp = walls + sector->wallptr;
	for (i = 0; i < sector->wallnum; i++) {
		ShowWall (wp, walls+(wp -> point2), 0);
		wp++;
	}

};


void ShowDirArrow (int x, int y, int ang) {
/* Shows an arrow which shows which way the sprite faces */
	double dir;
	int x2, y2;
	dir = Ang2Radians (ang);
	setcolor (15);
	setlinestyle (DOTTED_LINE, 0, 1);
	x2 = x + cos (dir) * Screen.spritewidth/2;
	y2 = y - sin (dir) * Screen.spritewidth/2;
	line (x, y, x2, y2);
	setlinestyle (0, 0, 0);
	line (x2, y2, x2 + cos (dir + 7 * PI / 8) * Screen.spritewidth/6,
		y2 - sin (dir + 7 * PI /8)* Screen.spritewidth/6);
	line (x2, y2, x2 + cos (dir - 7 * PI / 8) * Screen.spritewidth/6,
		y2 - sin (dir - 7 * PI /8)* Screen.spritewidth/6);
}


void ShowSprite (Tsprite *sprite, int clearing) {
long x1, y1; long swid, shei; int col; int thick=1;
int base=0;
	x1 = MapXScreenX (sprite->x);
	y1 = MapYScreenY (sprite->y);

	if (x1 > Screen.mapleft + 5 && x1 < Screen.mapright - 5 && y1 >
		Screen.maptop + 5 && y1 < Screen.mapbottom - 5) {
		col = Colors.plainsprite;

		if ((sprite->cstat.orientation)==1)  // Posters, flat things
			col = Colors.wallflat;
		if ((sprite->cstat).blocking)        // Solid sprites
			col = Colors.blockingsprite;

		if (sprite->picnum <= GPSPEED)// Sector Effectors, Music SFX, etc...
			col = Colors.se;
		if (sprite->picnum >= FIRSTGUNSPRITE && sprite->picnum <= BOOTS ||
			sprite->picnum == ATOMICHEALTH) //Items, weapons, ammo
			col = Colors.item;
		if (sprite->picnum >= LIZTROOP && sprite->picnum <= BOSS3+49) //Actors
			col = Colors.enemy;
		if (sprite->picnum == RAT) 				//rat
			col = Colors.rat;

		if ((sprite->cstat.orientation)==-2)  // Things on ground, flat
			col = Colors.groundflat;

		if (sprite->pal==BLUEPAL)                  // Multiplayer games only.
			col = Colors.netgames;
		if (sprite->picnum == APLAYER)
			col = Colors.playerstart;
		if (sprite->picnum == SEENINE)
			col = Colors.seenine;

		if (sprite->cstat.hitscan || sprite->picnum==SEENINE)
			thick =3;



		if (sprite->cstat.realcentering)
		base=1;
			//setlinestyle (DOTTED_LINE,  0, thick);

		else
			setlinestyle (SOLID_LINE,  0, thick);

		if (clearing) col = 0;
		setcolor (col);

		// Draw a cross to represent the sprite, just 2 lines in an X

		swid = Screen.spritewidth*sprite->xrepeat/100;
		shei = Screen.spritewidth*sprite->yrepeat/100;

		if (sprite->picnum == ACCESSCARD) {
			switch (sprite->pal) {
				case 0: setcolor (9); break;  // blue keycard
				case REDKEY: setcolor (4); break; // red keycard
				case YELLOWKEY: setcolor (14); break; // yellow keycard
			}
			setfillstyle (1, getcolor());
			bar (x1 - swid, y1-shei, x1+swid, y1+shei);
		}
		else {
			line (x1-swid, y1-shei,
				x1+swid, y1+shei);
			line (x1+swid, y1-shei,
				x1-swid, y1+shei);
		}
		if (sprite->lotag == 7 && sprite-> picnum == SECTOREFFECTOR){
			// Teleporter so surround with circle
			setcolor (Colors.teleporter);
			if (clearing)
			setcolor (0);
			circle (x1, y1, SPRITEWIDTH*100/Screen.scalex);
		}
		if (Screen.showspritedirs && Screen.scalex < 30)
			ShowDirArrow (x1, y1, sprite->ang);
		if (base) {
			setlinestyle (DOTTED_LINE,  0, thick);
			line (x1-swid*.9, y1+shei*.9,
				x1+swid*.9, y1+shei*.9);
		}

	}
}

void SpriteInfo (Tsprite *sprite, int spritenum) {
/* Shows some basic info about a selected sprite on screen */
char S [50]; char s[50]; char *s2;
	int d;
	setfillstyle (0,0);
	bar (Screen.mapleft, Screen.mapbottom+10, Screen.mapright, Screen.maxy);

	moveto (Screen.mapleft, Screen.mapbottom + 10);
	setcolor (11);
	memset (S, '\0', sizeof (S));

	itoa (spritenum, s, 10);

	strcat (S, "Sprite number ");
	strcat (S,s);
	strcat (S, "   Tile number ");

	itoa (sprite->picnum, s, 10);
	strcat (S, s);

	outtext (S);

	outtext ("  Hitag: ");

	ultoa (sprite->hitag, s, 10);
	outtext (s); outtext ("  Lotag: ");
	ultoa (sprite->lotag, s, 10);
	outtext (s);
	outtext ("   Name: ");
	FindTileName (element, sprite->picnum, &s2);
	outtext (s2);
	moveto (Screen.mapleft, Screen.mapbottom + 20);
	switch (sprite->picnum) {

		case SECTOREFFECTOR:  // SE's have types: find name.
			outtext ("SE type: ");
			outtext (GetSEName (sprite->lotag));
			break;
		case RESPAWN:
			FindTileName (element, sprite->hitag, & s2);
			if (s2 != NULL) {
				outtext ("Respawns a: ");
				outtext (s2);
			}
			break;
		case APLAYER:
			if (sprite->lotag == 1)
				outtext ("Co-op start point");
			else
				outtext ("Dukematch start point");
			break;

	}
	ultoa (sprite->sectnum, s, 10);
	outtext ("Sector number: ");
	outtext (s);
}

void WallStats (Twall *wall) {
//	assert (wall->picnum >=-1 && wall->picnum < 4096);
//	if (wall->nextsector > -1)
//		assert (!wall->cstat.masked);
	if (wall->picnum == 0) Stats.brownwalls++;
	if (wall->nextsector != -1) Stats.twosided++;
	if (wall->cstat.masked) Stats.maskedwalls++;
	if (wall->overpicnum == 560) Stats.mirrors++;
}

void SpriteStats (Tsprite *sprite) {
	// Constants from NAMES.H
	if (sprite->cstat.blocking) Stats.blocksprites++;
	if (sprite->cstat.orientation==1) Stats.flatsprites++;
	if (sprite->cstat.orientation==-2)  Stats.floorsprites++;
	if (sprite->picnum == CAMERA1) Stats.vidcameras++;
	if (sprite->picnum== VIEWSCREEN) Stats.vidscreens++;
	if (sprite->pal==1) Stats.netgames++;
	if (sprite->picnum==APLAYER) Stats.startpoints++;
	if (sprite->xrepeat > Stats.fatsprite) Stats.fatsprite=sprite->xrepeat;
	if (sprite->yrepeat > Stats.tallsprite) Stats.tallsprite=sprite->xrepeat;
	if (sprite->xrepeat < Stats.thinsprite) Stats.thinsprite=sprite->xrepeat;
	if (sprite->yrepeat < Stats.shortsprite) Stats.shortsprite=sprite->xrepeat;
	if (sprite->picnum==1 && sprite->lotag == 27) Stats.democams++;
}


void ShowStats (void) {
	printf ("****** STATISTICS ******\n");
	printf ("Number of sectors: %d\n", Stats.numsectors);
	printf ("Number of walls: %d\n", Stats.numwalls);
	printf ("Number of sprites: %d\n", Stats.numsprites);
	printf ("Highest floor: %ld\n", Stats.hifloor);
	printf ("Lowest floor: %ld\n", Stats.lofloor);
	printf ("Highest ceiling: %ld\n", Stats.hiceil);
	printf ("Lowest ceiling: %ld\n", Stats.loceil);
	printf ("Total of %d brown (texture 0) wall", Stats.brownwalls);
	printf ((Stats.brownwalls!=1) ? "s\n":"\n");   // print s for plural
	printf ("There are %d two-sided (red) walls (%d maskable)\n", Stats.twosided / 2, Stats.maskedwalls);
	// /2 because each 2 sided wall is really 2 walls on top of each other
	printf ("%d Video Screens connect to %d Video Cameras.  (%d Demo Cameras)\n",
	Stats.vidscreens,	Stats.vidcameras, Stats.democams);
	printf ("There are %d solid (blocking) sprites.\n", Stats.blocksprites);
	printf ("There are %d flat sprites vertically, and %d horizontally.\n",
		Stats.flatsprites, Stats.floorsprites);
	printf ("In netgames, there are %d start points, and %d extra sprites\n",
		Stats.startpoints, Stats.netgames);
	printf ("%d of the %d sectors are tilted\n", Stats.tiltsectors, Stats.numsectors);
	printf ("There are %d mirrors in this level\n", Stats.mirrors/2);
	printf ("Widest Sprite: %d  Thinnest Sprite: %d  Tallest Sprite: %d  Shortest Sprite: %d\n",
		Stats.fatsprite, Stats.thinsprite, Stats.tallsprite, Stats.shortsprite);

}

int LoadGraph (void) {
	int gm, gd;
//	gd = installuserdriver ("SVGA256", NULL);
	gd = VGA;
	gm = VGAHI;
	printf ("Mem at the moment, %ld", coreleft());
	initgraph (&gd, &gm, "");
	return graphresult();
}

void DrawWalls (Twall *wall, int numwalls, int walllevel) {
long i;
int draw;
	/* Draw wall by wall, to ensure every wall is drawn */
	for (i=0; i < numwalls; i++) {
//		delay (1);
		draw=1;
		if (walllevel == 0) draw = 1;
		if (walllevel == 1 && (wall+i)->nextsector!=-1) draw = 0;
		if (walllevel == 2) draw = 0;
		// Ttry and avoid drawing red walls twice. ## try later
		/*if (i > wall-> nextwall)*/
		if (wall[i].nextwall==-1 || !((wall + (wall+i)->nextwall)->cstat.oneway))
			if (draw)
				ShowWall (wall+i, wall+(wall+i)->point2, 0);

	}
}

void HideWalls (Twall *wall, int numwalls) {
int i;
	/* Draw wall by wall, to ensure every wall is drawn */
	for (i=0; i < numwalls; i++) {
		ShowWall ((wall)+i, wall+((wall)+i)->point2, 1);
	}
}

void DrawSprites (Tsprite *sprite, int numsprites, int spritelevel) {
int i;
int draw;
	/* Draw sprite by sprite, to ensure every sprite is drawn */
	for (i=0; i < numsprites; i++) {
		draw = 1;
		if (spritelevel <= 0) draw = 0;   // draw no sprites
		if (spritelevel <= 1) // draw any uncovered sprites
			if (((sprite+i)->picnum >= FIRSTGUNSPRITE &&
				(sprite+i)->picnum <= BOOTS) ||
				(sprite+i)->picnum >= LIZTROOP &&
					(sprite+1)->picnum <= BOSS3+49 ||
				(sprite+i)->picnum == ATOMICHEALTH)
				draw = 0;
		if ((sprite+i)->cstat.orientation) // draw items and monsters
			if (spritelevel <= 2)
				draw = 0;

		if ((sprite+i)->pal == 1 || (sprite+i)->picnum== RAT)
													// draw multiplayer too and rats
			if (spritelevel <= 3)
				draw = 0;
		if ((sprite+i)->picnum < 10 || (sprite+i)->picnum == SEENINE)
			// draw se's and explosives
			if (spritelevel <= 4)
				draw = 0;
		if (spritelevel==5)					  // draw everything
			draw = 1;
		if (draw)
			ShowSprite (sprite + i, 0);

	}
}

void HideSprites (Tsprite *sprite, int numsprites) {
int i;
	for (i=0; i < numsprites; i++) {
		ShowSprite (sprite + i, 1);
	}
}


int LoadMap (char * filename, int *handle, Tmap *map) {
// Loads a duke3d map file from disk, into the version, playerstart, sectors
// walls and sprites

	*handle = open (filename, O_RDONLY);
	if (*handle <= 0) {
		printf ("Can't find file: %s", filename);
		exit (3);
	}

	FileRead (*handle, &map->version, sizeof (map->version));
	FileRead (*handle, &map->playerstart, sizeof (map->playerstart));

	FileRead (*handle, &map->numsectors, sizeof (map->numsectors));
//	assert (map->numsectors >= 0 && map->numsectors <= 1024);
	if (!(map->sector = FPAllocMem ((long)sizeof (Tsector)*(map->numsectors)))) {
		printf ("Not enough mem to load sectors...aborting.");
		exit (1);
	};
	FileRead (*handle, map->sector, sizeof (Tsector)*(map->numsectors));

	FileRead (*handle, &map->numwalls, sizeof(map->numwalls));
//	assert (map->numwalls >= 0 && map->numwalls <= 8192);
	if (!(map->wall= FPAllocMem ((long)sizeof (Twall) * (map->numwalls)))) {
		printf ("Not enough mem for walls...aborting\n");
	}
	FileRead (*handle, map->wall, (long)sizeof (Twall)*(map->numwalls));

	// ici se trouve le probleme...
	FileRead (*handle, &map->numsprites, sizeof(map->numsprites));
//	assert (map->numsprites >= 0 && map->numsprites <= 4096);
	if (!(map->sprite = FPAllocMem ((long)sizeof (Tsprite) * (map->numsprites)))) {
		printf ("Not enough mem for sprites...aborting");
	}
	FileRead (*handle, map->sprite, sizeof (Tsprite)*(map->numsprites));

	return 0;
}

int SaveMap (char * filename, Tmap *map) {
/* Savemap does the opposite of loadmap - takes a map from memory and writes
it to the filename provided.  Not it doesn't return a handle, because you
don't need one */
int handle;

	handle = open (filename, O_RDONLY); // Check if file exists already
	if (handle <= 0);
		// File hasn't been used before - Good!
	else
		// Do something because we don't want to accidentally write over files
	;


	close (handle);
	handle = creat (filename, S_IWRITE);
	if (handle <= 0) {
		printf ("Can't find file: %s", filename);
		exit (3);
	}

	FileWrite (handle, &map->version, sizeof (map->version));
	FileWrite (handle, &map->playerstart, sizeof (map->playerstart));

	FileWrite (handle, &map->numsectors, sizeof (map->numsectors));
	FileWrite (handle, map->sector, sizeof (Tsector)*(map->numsectors));

	FileWrite (handle, &map->numwalls, sizeof(map->numwalls));
	FileWrite (handle, map->wall, sizeof (Twall)*(map->numwalls));

	FileWrite (handle, &map->numsprites, sizeof(map->numsprites));
	FileWrite (handle, map->sprite, sizeof (Tsprite)*(map->numsprites));

	return 0;
}


void DisplayMap (Tmap *map) {
long i=0;
long j=0;
int lastwall, curwall;
		if (!Screen.Usinggraph) {
			for (i=0; i < map->numwalls; i++) {
				printf ("Wall %ld -> wall %d\n", i, (map->wall + i)->point2);
				gotoxy (1,1);
				printf ("Wall %d's ", i);
	//			ShowWallTxt (wall + i);
			}
		}
		else {
			switch (Screen.drawmethod) {
				case SECTORS: {
					/* Use the sector by sector approach for drawing walls */
					for (i=0; i < map->numsectors; i++) {
						lastwall = (map->sector+i) -> wallptr;
						for (j=0; j<(map->sector+i)->wallnum; j++) {
							curwall= (map->wall + lastwall)->point2;
							ShowWall (map->wall + lastwall, map->wall + curwall, 0);
							lastwall = curwall;
						}
					}
					break;
					}
				case WALLS: {
					DrawWalls (map->wall, map->numwalls, Screen.walllevel);
				break;
				}
			}
			DrawSprites (map->sprite, map->numsprites, Screen.spritelevel);
			ShowPlayerStart(&(map->playerstart), 0);
		}

}

void DrawScreen (Tmap *map) {
/* Completely draws screen and map and anything else needed */
	setcolor (7);
	setlinestyle (0,0,1);
	rectangle (Screen.mapleft, Screen.maptop, Screen.mapright, Screen.mapbottom);
	DisplayMap (map);
	setcolor (9);
	setlinestyle (0,0,1);
	circle ((Screen.mapleft + Screen.mapright)/2,
		(Screen.maptop + Screen.mapbottom) / 2, 5); // Circle in middle
	if (Cursprite >= 0) {
		HiLiteSprite (map->sprite+Cursprite);
	}
}

void ClearScreen(Tmap *map) { // Using one method or other, makes screen black.
	if (Screen.Usinggraph)
		if (Screen.clearmethod) {
			HideWalls(map->wall, map->numwalls);
			HideSprites (map->sprite, map->numsprites);
			ShowPlayerStart(&(map->playerstart), 1);
		}
		else
			cleardevice();
	else
		if (Screen.Usinggraph);
}

int SelectSprite (Tmap *map) {
/* Choosese a sprite and visually represents it */
/* Returns if this is a different sprite than before */
int old;
	old = Cursprite;
	Cursprite = GetSpriteNum (ScreenXMapX (Mousex),
		ScreenYMapY (Mousey), map);
	if (Cursprite != old) {
		HideMouse ();
		if (old > -1) {
			UnHiLiteSprite (map->sprite + old);
			ShowSprite (map -> sprite + old, 0);
			HideSector (map->sector + map->sprite [old].sectnum,
				map->wall);
		}
		if (Cursprite >= 0) {
			HiLiteSprite(map->sprite + Cursprite);
			ShowSprite (map ->sprite + Cursprite, 0);
		}
		ShowMouse ();
	}
	if (old != Cursprite && Cursprite >= 0)
		return 1;
	else
   	return 0;
}

void CentreMap (void) {
	Screen.centrex = ScreenXMapX (Mousex);
	Screen.centrey = ScreenYMapY (Mousey);
}


int GetInput (void) {
/* Reads from keyboard and mouse to give a code for what user did */
/* Possible codes returned from GetInput:
	0-255: Ascii character of key ptressed.
	256 - 511: Extended character such as arrows.
   1000 - 10? : Mouse action such as button release */

static int oldmouseb=0; // Old state of mouse buttons used for mouse release
int code = 0;

	oldmouseb = Mouseb;
	GetMouseState();
	if ((Mouseb==0 && (oldmouseb) && Dragmap == 0)
		|| (Mouseb == 1 || Mouseb == 2) && Dragmap == 1) {
			//mouse released - give special mouse code
		if (oldmouseb == 1) code = 1001;
		if (oldmouseb == 2) code = 1002;
		if (oldmouseb == 3)
			code = 1003; /* Both mouse buttons released */

	}
	if (kbhit()) {
		code = getch();
		if (code == 0) code = 256 + getch();
	}
	if (code == 0) code = 1001;
	return code;
}

void MainLoop (Tmap *map) {
int done=0;
int k=0;
int redraw=0; // Update the screen after this keypress?
	DrawScreen (map);
	while (!done) {

		redraw = 0;
		k = GetInput();
		if (k) {
			if (Screen.clearmethod) {
				HideMouse();
				ClearScreen (map);
			}
		}
		switch (k) {
			case 27: done = 1; break;
			case 256+72: Screen.centrey -= 1000*Screen.scaley/100; redraw++;break;
			case 256+75: Screen.centrex -= 1000*Screen.scalex/100; redraw++;break;
			case 256+77: Screen.centrex += 1000*Screen.scalex/100; redraw++;break;
			case 256+80: Screen.centrey += 1000*Screen.scaley/100; redraw++;break;
			case 1001:
				if (Screen.selectsprites) if (SelectSprite (map)) {
					SpriteInfo (map->sprite + Cursprite, Cursprite);
					ShowSector (map->sector + map->sprite [Cursprite].sectnum,
						map->wall);
				}
				break;
			case 1002:
				CentreMap ();
				if (!Dragmap) {
					if (_bios_keybrd(_KEYBRD_SHIFTSTATUS) & 0x3) // L+R shift
						ZoomOut();
					else
						ZoomIn();
				}
				redraw++;
				break;
			case 1003: CentreMap (); redraw=1;break;
			case '-': ZoomOut (); redraw++; break;
			case '+':
			case '=': ZoomIn(); redraw++; break;
			case 's': Screen.spritelevel = (++Screen.spritelevel) % (MAXSPRITELEVEL+1);
				gotoxy (1,1);
				printf ("Sprite level %d", Screen.spritelevel);
				redraw++;
			  break;
			case 'w': Screen.walllevel = (++Screen.walllevel) % (MAXWALLLEVEL+1);
				gotoxy (1,1);
				printf ("Wall level %d", Screen.walllevel);
				redraw++;
			  break;
			case 'M': InitMouse(); break;
			case '/': case '?':
				SelectSprite (map);
				/* Deliberate fall-through */
			case 'i':
				if (Cursprite >= 0) SpriteInfo (map->sprite + Cursprite, Cursprite);
				break;
			case 'd':
				Dragmap = 1-Dragmap; break;
			case '|':
				printf (""); break;
			case '8':
				redraw++;
				CreateSprite (map, 1, 181, 14, Screen.centrex, Screen.centrey);
			break;
			case 'm':
				redraw++;
				Screen.clearmethod = !Screen.clearmethod;
				setcolor (7);
				if (Screen.clearmethod)	{
					outtextxy (1,1, "Clear method on!");
					ClearScreen (map);
				}
				break;
			case '*': /*Start Mouse drawing */
				moveto (Mousex, Mousey);
				setlinestyle (0, 0, 3);
				setcolor (14);
			break;
			case 'S':
				Screen.selectsprites = 1 - Screen.selectsprites;
         break;
		}
		if (Mouseb == 1) {
			HideMouse ();
			lineto (Mousex, Mousey);
			ShowMouse ();
		}
		if (k) {
			if (redraw || Screen.drawmethod) {
			if (!Screen.drawmethod) HideMouse();
			ClearScreen(map);
			DrawScreen (map);
			}
			ShowMouse();
		}
	}
}

void InitScreen (Tplayerstart *pstart) {
int result;
	if (Screen.Usinggraph) {
		result=LoadGraph ();
		if (result) {
			printf ("Problem loading graphics driver - ");
			printf ("%s", grapherrormsg (result));
			exit (2);
		}
		else {
			Screen.maxx = getmaxx ();
			Screen.maxy = getmaxy ();
		};
	}
	Screen.centrex = pstart->posx;
	Screen.centrey = pstart->posy;
	Screen.mapleft = 5;
	Screen.maptop = 5;
	Screen.mapright = Screen.maxx - 5;
	Screen.mapbottom = Screen.maxy -100;

	Screen.scalex = 50;
	Screen.scaley = 50;
	Screen.showspritedirs = 1;
   Screen.selectsprites = 1;
	ZoomIn(); ZoomOut();
	Colors.solidwall = WHITE;  Colors.doublewall=RED;
	Colors.mirror = LIGHTGREEN;	Colors.maskable = LIGHTBLUE;
	Colors.blockingwall = LIGHTMAGENTA; Colors.oneway = YELLOW;

	Colors.item = LIGHTGREEN;  Colors.se = YELLOW;  Colors.seenine = BROWN;
	Colors.rat = CYAN;     Colors.netgames = BLUE;
	Colors.playerstart = LIGHTBLUE;
	Colors.groundflat = WHITE;  Colors.wallflat = LIGHTGREY;
	Colors.enemy = LIGHTRED;
	Colors.blockingsprite = LIGHTMAGENTA;  Colors.plainsprite = LIGHTCYAN;
	Colors.sector = LIGHTGREEN;  Colors.spritebox = LIGHTRED;
	Colors.teleporter = LIGHTCYAN;

}

void InitStats (Tmap *map) {
	Stats.thinsprite = 1000; Stats.fatsprite = 0;
	Stats.shortsprite = 1000; Stats.tallsprite = 0;
	Stats.lofloor=map->sector->floorz;	Stats.hifloor=map->sector->floorz;
	Stats.loceil=map->sector->ceilingz;	Stats.hiceil=map->sector->ceilingz;
	Stats.numsectors = map->numsectors;
	Stats.numwalls = map->numwalls;
	Stats.numsprites = map->numsprites;
}

void Init (int argc, char *argv [], Tmap **map) {
	char *filename;
	long i= 0;

	Screen.Usinggraph = 1;

	clrscr();
	if (Verbose||1)
		printf ("Core left: %ld Kb, %ld bytes",coreleft()/1024, coreleft ());
	if (argc < 2)
		filename = "e1l1.map";
	else
		filename = argv [1];

	element = LoadNameH();

	if (Verbose) printf ("Showing %s\n", filename);
	*map = FPAllocMem (sizeof (Tmap));
	LoadMap (filename, &mapfile, *map);
	InitStats (*map);

	InitScreen(&((*map)->playerstart));

	for (i=0; i < (*map)->numsectors; i++) SectorStats ((*map)->sector + i);
	for (i=0; i < (*map)->numwalls; i++) WallStats ((*map)->wall + i);
	for (i=0; i < (*map)->numsprites; i++) SpriteStats ((*map)->sprite + i);
	InitMouse();
	ShowMouse();
	Screen.drawmethod = WALLS;
	Screen.spritelevel = MAXSPRITELEVEL;
	Screen.walllevel = 0;

}

void Exec (Tmap *map) {



	printf ("\n");


	MainLoop (map);
	if (Screen.Usinggraph) closegraph();
	ShowStats();
//	SaveMap ("tmap.map", map);
	close (mapfile);
}

void Terminate (Tmap *map) {
	FPFreeMem (map->sector);
	FPFreeMem (map->wall);
	FPFreeMem (map->sprite);
	FPFreeMem (map);
	FreeNamedTiles(element);
	printf ("Core left now is %ld", coreleft ());
	getch ();
}

void main (int argc, char *argv []) {

	Tmap *map;
	Init (argc, argv, &map);
	Exec(map);
	Terminate (map);
//	getch();
}